;
; On-board flash programmer on MachZ BUR environment for Intel E28F128 chip. This programmer
; works likely with other CFI devices as well but has not been tested with any other
; chips than Strataflash.
;
; This programmer writes variable-length payload into flash device.
;
; Programmer takes two parameter records at ZTAG manager software:
; 	- Start address in flash
;	- Erase flag
;
; Programmer will program entire payload to flash chip, starting from defined
; address, without checking the overruns. If the chip is smaller than needed, wrap-around
; will occure.
;
; (c)2000 ZF Linux Devices, Inc.
;
; Target Chip:	CFI flash chip (Intel Strataflash)
; Size:		16M
; Chipselect:	ms_cs0
; Mode:		autodetect by chip connection
; Chip page:	defined by Parameter record
; Window base:	E0000h
; Window size:	10000h (64K)
;
;
; This code executes as BUR "Load and execute" function. It will fetch parameter
; record and payload code following to the executable code in dongle.
;
; Compile using MASM 6.11 with following command line:
;
;	ml /nologo /Sc /Fl /Zm /Fm /Ta E28F128.ASM
;
; The resulting .COM file can be directly used as payload for 
; BUR "Uplaode and Execute" function
;
;

.MODEL TINY
.CODE
.486p

PRGBASE		=	0E0000h			; where to create programming window

		include FLASHPGM.MAC

		org	0
START:			
		push	cs
		pop	es
		
		call	CRLF
		SAYCRLF	'PGM - MachZ Common Flash Interface Programmer V1.12'
		SAYCRLF	'-----------------------------------------------------'
				
		
		call	ZTPrepareRead
		
		; Fetch base address
		
		call	ParmRecord2EAX		; fetch parameter record to EAX
		jnc	Parm1OK

		SAYCRLF	'Flash programming start address missing!'		
		jmp	ExitPgmrFail				
Parm1OK:
		SAY	'Programming starts from 0x'		
		call	SerOut32
		call	CRLF
		
		and	eax, 0FFFFF000h
		mov	cs:[FlashBase], eax
		
		; Fetch Erase parameter
		
		call	ParmRecord2EAX		; fetch parameter record to EAX
		jnc	Parm2OK

		SAYCRLF	'Sector erase parameter missing!'
		jmp	ExitPgmrFail				
Parm2OK:
		mov	cs:[EraseSector], eax

		; Go on with basket now		
		
		PROCESS_BASKET_HEADER	cs:[BasketSize], ExitPgmrFail		; fetch basket header and display info from
		
						
		; OK, bells and whistles are there. Do programming now.
		
		MEMWIN	CS0, _16BIT_RW, PRGBASE, (64*1024), cs:[FlashBase]
						
		; Turn ISA bus speed as slow as possible, since we do not know what the 
		; SYSCLK is and we can not program device well if we do it too fast
		
		SLOW_ISA DIV_8			; divide ISA speed by 8
		
		; Now we have chip layed out from E000:0 to E000:FFFF
		; Time to program some flash.
		
		; Check, whenever or not we have CFI-compliant part at chipselect
		
		push	(PRGBASE/10h)
		pop	ds		
		xor 	si, si
		mov	byte ptr ds:[si], 98h	; CFI "Read Query" command
		
		mov	eax, ds:[si+20h]
		mov	ebx, eax		; store temporarily
		and	eax, 00FF00FFh		; strip, in case it was connected as 8-bit device
		cmp	eax, 00520051h		; check, if the chip responded with "Q.R."
		jz	@f
QueryFailed:
		SAYCRLF	'FAILED: Flash device did not respond to Read Query command'
		jmp	ExitPgmrFail	
@@:
		cmp	ebx, 0FF52FF51h		; check, if the Flash is strapped to 8-bit mode ('Q'/xff'R'/xff)
		jnz	CFI_OK
		
		MEMWIN_MODE CS0, _8BIT_RW	; switch chipselect over to 8-bit mode
				
		mov	ebx, ds:[si+20h]
		cmp	ebx, 'RRQQ'
		jnz	QueryFailed		; wrong answer ...

		SAYCRLF	'Flash chip is strapped to 8-bit width'
		
		mov	word ptr cs:[_8BitMode], 1
CFI_OK:
		; Looks like we have CFI device connected.
												
		SAY	'CFI compliant Flash device: MfgID='
		
		mov	al, ds:[si+0]
		call	SerOut8
		
		SAY	' DevID='
		
		mov	al, ds:[si+2]
		call	SerOut8
		call	CRLF
				
		;
		; Check, if the device supports write buffer
		
		mov	cl, ds:[si+(2*2Ah)]	; buffer length is 2^bl bytes
		cmp	cl, 0
		jz	NoWB			; do not support write buffer if only 1 bytes long
		
		SAY	'Device Write Buffer length is 0x'
		
		mov	ax, 1
		shl	ax, cl
		call	SerOut16
		SAYCRLF	' bytes'
		
		cmp	ax, 32
		jbe	@f
		mov	ax, 32			; strip write buffer to 32 bytes if longer		
@@:
		mov	cs:[WBSupported], al
NoWB:
		mov	al, 0FFh
		mov	ds:[si], al		; end query mode
		
		; ====== Basket handling begins here ======
		;
		; Note, that in order to maintain correct checksum for basket, 
		; we must calculate data CRC for RLE header as well!
		;
		
		call	ResetCRC
		
		PROCESS_RLE_HEADER  ExitPgmrFail		; process RLE header now		
		
		mov	ebp, dword ptr cs:[BasketSize]
		
		cmp	word ptr cs:[_8BitMode], 1
		jz	SizeOK			; basket size can be an odd number if we program 8-bit baskets
		
		test	ebp, 1			; odd bytecount in basket?
		jz	SizeOK
		
		SAYCRLF	'Word programming requires Even payload size!'
		jmp	ExitPgmrFail			
SizeOK:
		
PrgLoop:
		; We seem to have right chip. Erase sector now, if requested by parameter
		
		SAY	'0x'
		
		mov	eax, cs:[FlashBase]
		
		call	SerOut32
		SAY	': '		
		
		; Check, if erase is neccessary
		
		cmp	dword ptr cs:[EraseSector], 1
		jnz	SectorEmpty
		
		; Erase only sectors, what are 128K aligned. Skip this check, however, if
		; it is first sector we are writing.
		
		cmp	byte ptr cs:[FirstPass], 1
		jz	@f
		
		test	dword ptr cs:[FlashBase], 10000h
		jnz	SectorEmpty
@@:
		mov	byte ptr cs:[FirstPass], 0
		
		SAY	'Erasing .. '
				
		push	(PRGBASE/10h)
		pop	fs
		xor	si, si
		
		mov	byte ptr fs:[si], 20h	; request block erase
		jmp	$+2
		mov	byte ptr fs:[si], 0D0h	; confirm
		
		; Erase is now in progress. Check, when we are ready.
		
		call	DSBX2Var
		mov	ax, 182	
		mov	ds:[bx.CountDown], ax	; gives 10sec. timeout
		
		mov	dx, 218h
		mov	al, 05Eh ;ZT_SIG_OUT
		out	dx, al
		inc	dx
		in	al, dx
		and	al, 11110101b			; disable LED's		
		out	dx, al 
@@:
		cmp	word ptr ds:[bx.CountDown], 0
		jz	@f		
		test	byte ptr fs:[si], 10000000b
		jz	@b				; loop until erase completes
@@:
		mov	word ptr ds:[bx.CountDown], 0	; reset timer, so we will not loose our blinking LED's
		cmp	byte ptr fs:[si], 80h
		jz	@f
		
		SAYCRLF	'FAILED!'
		SAY	'Flash Error Code = 0x'
		
		mov	al, fs:[si]
		call	SerOut8
		call	CRLF
		
		jmp	ExitPgmrFail
@@:		
		mov	byte ptr fs:[si], 0FFh		; back to array mode
SectorEmpty:		
		; All set. Programming ...
		
		call	ZTPrepareRead			; ZFLogic back to track
		
		SAY	'Programming .. '		
		
		mov	ecx, 64*1024
		cmp	ecx, ebp
		jbe	@f
		mov	ecx, ebp		; if less than 64K left, program it		
@@:
		cmp	word ptr cs:[_8BitMode], 1
		jz	@f
		shr	ecx, 1			; set ECX to word count
@@:
		cmp	ecx, 0
		jz	PgmDone			; don't do anything if basket size is 0		
		
		push	(PRGBASE/10h)
		pop	ds
		xor	si, si
		
BytePrgLoop:		
		mov	byte ptr ds:[si], 40h
		call	ZTRead_RLE			
		cmp	word ptr cs:[_8BitMode], 0
		jz	Prg16Bit
		
		mov	byte ptr ds:[si], al
		mov	bl, al				; save for vrfy
		jmp	Prg8_16Write
Prg16Bit:
		shl	ax, 8
		call	ZTRead_RLE
		xchg	al, ah
		mov	ds:[si], ax
		mov	bx, ax				; save for vrfy
		
Prg8_16Write:
				
		; on the extreme case we can have PCI backside clock 80Mhz. ISA
		; divider is 8, so we have 10M ISA bus clock. Programming cycle can be max.
		; 50us, so we need to wait here about 500 ISA cycles to kill that time.
		
		mov	di, 500
WaitWriting:
		cmp	byte ptr ds:[si], 80h	; see if "ready" bit is set
		jz	ByteOk
		in	al, 80h				; create one ISA cycle for delay
		dec	di		
		jnz	WaitWriting
		
		jmp	PgmFailed		
ByteOk:				
		cmp	word ptr cs:[_8BitMode], 1
		jnz	@f
		mov	byte ptr ds:[si], 0FFh		; go to array mode for 8-bit device
		jmp	ArrayModeOk		
@@:
		mov	word ptr ds:[si], 0FFFFh	; go to array mode for 16-bit device
ArrayModeOk:				
		cmp	bl, ds:[si]
		jnz	PgmFailed
			
		cmp	word ptr cs:[_8BitMode], 1
		jz	@f

		cmp	bx, ds:[si]
		jnz	PgmFailed
@@:
		add	si, 2
		sub	si, word ptr cs:[_8BitMode]	; decrement for byte count according to byte mode ...
		
		loopd	BytePrgLoop		
PgmDone:
		SAYCRLF	'OK!'
				
		; Anything left?
		
		sub	ebp, 64*1024
		jz	@f
		jc	@f
				
		add	dword ptr cs:[FlashBase], 64*1024	; update page to next
		
		MEMWIN_FLASHBASE CS0, cs:[FlashBase]
		
		jmp	PrgLoop
@@:
		mov	byte ptr ds:[si], 0FFh				; back to array mode
		
		; Now when programming is done, go check the data checksum if it was RLE image
		
		cmp	byte ptr cs:[BasketType], 0FFh
		jz	@f
		
		mov	eax, dword ptr cs:[RLE_CheckSum]
		cmp	eax, dword ptr cs:[RLE_Chk]		; compare against calculated checksum
		jz	@f
		
		SAY	'RLE basket checksum error '
		
		call	SerOut32
		mov	al, ' '
		call	SerSend2
		mov	eax, dword ptr cs:[RLE_Chk]
		call	SerOut32	
		call	CRLF
		
		jmp	ExitPgmrFail
@@:
		call	ZTPrepareRead
		
		; Get original checksum
		
		call	ZTRead
		shl	ax, 8
		call	ZTRead
		xchg	al, ah
		
		call	DSBX2Var		; get variables block to DS:BX
		cmp	word ptr ds:[bx.YModemCRChi_C], ax
		jz	CRCOK
	
		SAY	'Data CRC failure: '
		
		call	SerOut16
		mov	al, ' '
		call	SerSend2
		mov	ax, word ptr ds:[bx.YModemCRChi_C]
		call	SerOut16
		call	CRLF		
		jmp	ExitPgmrFail		
CRCOK:
		SAYCRLF	'Data CRC was OK!'		
		jmp	ExitPgmrOk
		
PgmFailed:		
		SAYCRLF	'FAILED!'
		SAY	'Flash Error Code = 0x'
		
		mov	al, ds:[si]
		call	SerOut8
		call	CRLF
				
ExitPgmrFail:
		; Lit up RED LED and do not do anything else!
		
		mov	dx, 218h
		mov	al, 05Eh ;ZT_SIG_OUT
		out	dx, al
		inc	dx
		in	al, dx
		and	al, 11110101b
		or	al, ZT_LED_RED		;00001000b
		out	dx, al	
		
		SAYCRLF	'Freezing because of Programming failure ...'
		
		jmp	$					
ExitPgmrOk:
		; Lit up GREEN LED and continue with BUR		
		mov	dx, 218h
		mov	al, 05Eh ;ZT_SIG_OUT
		out	dx, al
		inc	dx
		in	al, dx
		and	al, 11110101b
		or	al, ZT_LED_GREEN	;00000010b
		out	dx, al
ExitPgmr:
		call	CRLF
		
		; Set timer to maximum value. This is useful to prevent
		; BUR from blinking with LED's when loading next commands.
		; This way we maintain our RED or GREEN LED setting
		
		call	DSBX2Var
		mov	word ptr ds:[bx.CountDown], 0FFFFh
		
		; Always exit with ZFL registers prepared for accelerated read!	
			
		call	ZTPrepareRead		
				
		retf				; resume with BUR
		
		
		include BURAPI.ASM	; entry points to BUR
		include FLASHPGM.ASM	; flash programmer common routines

FlashBase	dd	?
_8BitMode	dw	0
EraseSector	dd	0
FirstPass	db	1

WBSupported	db	0		; if 0, write buffer is not supported. 
					; Any other value represents the write buffer length.
					
VirtualWB	db	32 dup(0)	; virtual write buffer
VirtualWB_bytes	db	0		; bytes in virtual write buffer

;
; Write down data gathered into virtual Write Buffer in correct mode 
;

WB_Flush	proc

		ret
WB_Flush	endp


;
; This function collects the single bytes to virtual vrite buffer and writes them down if
; write buffer is full.
;
; Entry:	AL - byte to write
;

Write_WB	proc

		ret
WRITE_WB	endp


END START
ENDS